unit fppdfindexer;

{$mode ObjFPC}{$H+}

{ $DEFINE USEFIREBIRD}
{$DEFINE USEPG}

{$IFNDEF USEPG}
  {$IFNDEF USEFIREBIRD}
      {$DEFINE USEPG}
  {$ENDIF}
{$ELSE}
  {$UNDEF USEFIREBIRD}
{$ENDIF}

interface

uses
  Classes, SysUtils, Types, fpindexer, sqldbindexdb , fppdfparser, fppdfobjects, contnrs;

Const
  PDFIndexDefaultMinWordLength = 3;

Type
  { TPDFIndexer }
  TPDFIndexLogEvent = procedure(Sender : TObject; Const Msg : String) of Object;

  TPDFIndexer = Class(TComponent)
  private
    FIgnoreWords: TStrings;
    FIndexDB: TSQLDBIndexDB;
    FLanguage: string;
    FMinWordLength: Integer;
    FOnLog: TPDFIndexLogEvent;
    FPDFURL: String;
    FHash : TFPStringHashTable;
    procedure DoIgnoreChange(Sender: TObject);
    procedure SetIgnoreWords(AValue: TStrings);
  Protected
    Procedure DoLog(const Msg : String);
    Procedure DoLog(const Fmt : String; Args : Array of Const);
    procedure DoIndexPDFPage(const aPage: TPDFPageObject; aPageNo: Integer); virtual;
    Procedure IndexPDFPage(const aPage : TPDFPageObject; aPageNo : Integer);  virtual;
    Procedure RebuildHash; virtual;
    Procedure DoIndexPDF(const aStream : TStream); virtual;
    function AllowedWord(aWord: String): Boolean; virtual;
    Function DoSplit(const aText : RawbyteString) : TStringDynArray; virtual;
  Public
    Constructor Create(aOwner : TComponent); override;
    Destructor Destroy; override;
    Procedure Connect;
    Procedure CreateTables;
    Procedure Disconnect;
    Procedure IndexPDF(const aFileName : string); overload;
    Procedure IndexPDF(const aStream : TStream); overload;
    Property PDFURL : String Read FPDFURL Write FPDFURL;
    Property Language : string read FLanguage Write FLanguage;
    Property Indexer : TSQLDBIndexDB Read FIndexDB;
    Property MinWordLength : Integer Read FMinWordLength Write FMinWordLength;
    Property IgnoreWords : TStrings Read FIgnoreWords Write SetIgnoreWords;
    Property OnLog : TPDFIndexLogEvent Read FOnLog Write FOnLog;
  end;

implementation

uses
{$IFDEF USEFIREBIRD}
  fbIndexdb
{$ENDIF}
{$IFDEF USESQLITE}
  sqliteindexdb
{$ENDIF}
{$IFDEF USEPG}
  pgIndexdb
{$ENDIF}
  ;
{ TPDFIndexer }

procedure TPDFIndexer.SetIgnoreWords(AValue: TStrings);
begin
  if FIgnoreWords=AValue then Exit;
  FIgnoreWords.Clear;
  FIgnoreWords.AddStrings(AValue);
end;

procedure TPDFIndexer.DoLog(const Msg: String);
begin
  If Assigned(FOnLog) then
    FOnLog(Self,Msg);
end;

procedure TPDFIndexer.DoLog(const Fmt: String; Args: array of const);
begin
  DoLog(Format(Fmt,Args));
end;

procedure TPDFIndexer.RebuildHash;

Var
  S : String;

begin
  FHash.Clear;
  For S in FIgnoreWords do
    FHash.Add(S,S);
end;

procedure TPDFIndexer.DoIgnoreChange(Sender: TObject);
begin
  RebuildHash;
end;

procedure TPDFIndexer.IndexPDFPage(const aPage: TPDFPageObject; aPageNo: Integer);

begin
  Indexer.BeginTrans;
  try
    DoIndexPDFPage(aPage,aPageNo);
    Indexer.CommitTrans;
    DoLog('Indexed page %d',[aPageNo])
  except
    On E : exception do
      DoLog('Error %s while indexing page %d: %s',[E.ClassName,aPageNo,E.Message]);
  end;
end;

procedure TPDFIndexer.DoIndexPDFPage(const aPage: TPDFPageObject; aPageNo: Integer);

Var
  aData: TSearchWordData;
  aCmd : TPDFCommand;
  aWord,FontName,Rawtext : RawByteString;
  aFontRef : TPDFRefData;
  UnicodeMap : TPDFCMap;
  aFontObj : TPDFFontObject;
  aDoc : TPDFDocument;

begin
  aDoc:=aPage.Document;
  With aData do
    begin
    FileDate:=Date;
    Language:=FLanguage;
    Position:=aPageNo;
    URL:=PDFURL;
    end;
  for aCmd in aPage.CommandList do
    begin
    if aCmd is TPDFTfCommand then
      begin
      FontName:=TPDFTfCommand(aCmd).FontName;
      if (FontName<>'') and (FontName[1]='/') then
        Delete(FontName,1,1);
      aFontRef:=aPage.FindFontRef(FontName);
      aFontObj:=aDoc.FindFont(aFontRef); // TPDFFontObject
      if Assigned(aFontObj) then
        UnicodeMap:=aFontObj.UnicodeCMap
      else
        UnicodeMap:=nil;
      end
    else If aCmd is TPDFTextCommand then
      begin
      rawText:=TPDFTextCommand(aCmd).GetFullText(UnicodeMap);
      aData.Context:=Rawtext;
      SetCodePage(RawText,CP_UTF8);
      for aWord in DoSplit(RawText) do
        begin
        aData.SearchWord:=aWord;
        Indexer.AddSearchData(aData);
        end;
      end;
    end;
end;

procedure TPDFIndexer.DoIndexPDF(const aStream: TStream);

var
  aPageNo : integer;
  aPage : TPDFPageObject;
  aParser : TPDFParser;
  Doc : TPDFDocument;

begin
  aPageNo:=0;
  Doc:=Nil;
  aParser:=TPDFParser.Create(aStream);
  try
    Doc:=TPDFDocument.Create();
    aParser.ResolveToUnicodeCMaps:=True;
    aParser.ParseDocument(Doc);
    For aPage in Doc.Pages do
      begin
      Inc(aPageNo);
      IndexPDFPage(aPage,aPageNo);
      end;
  finally
    doc.free;
    aParser.Free;
  end;
end;

function TPDFIndexer.AllowedWord(aWord : String) : Boolean;

begin
  Result:=Length(aWord)>MinWordLength;
  if Result then
    Result:=FHash.Find(aWord)=Nil;
end;

constructor TPDFIndexer.Create(aOwner: TComponent);
begin
  inherited Create(aOwner);
  FLanguage:='EN';
  FIgnoreWords:=TStringList.Create;
  FHash:=TFPStringHashTable.Create;
  FMinWordLength:=PDFIndexDefaultMinWordLength;
  TStringList(FIgnoreWords).OnChange:=@DoIgnoreChange;
  TStringList(FIgnoreWords).Sorted:=True;
  TStringList(FIgnoreWords).Duplicates:=dupIgnore;
{$IFDEF USEPG}
  FIndexDB:= TPGIndexDB.Create(Self);
{$ENDIF}
{$IFDEF USEFIREBIRD}
  FIndexDB:= TFBIndexDB.Create(Self);
{$ENDIF}
end;

destructor TPDFIndexer.Destroy;
begin
  FreeAndNil(FIndexDB);
  FreeAndNil(FIgnoreWords);
  FreeAndNil(FHash);
  inherited Destroy;
end;

procedure TPDFIndexer.Connect;
begin
  Indexer.Connect;
  DoLog('Connected to index database %s',[Indexer.DatabasePath]);
end;

procedure TPDFIndexer.Disconnect;
begin
  Indexer.DisConnect;
  DoLog('Disconnected from index database %s',[Indexer.DatabasePath]);
end;

function TPDFIndexer.DoSplit(const aText: RawbyteString): TStringDynArray;

Var
  aCount : Integer;
  c : AnsiChar;
  Cleaned : String;

  Procedure MaybeAdd;

  begin
    if (Cleaned<>'') and AllowedWord(Cleaned) then
      begin
      Result[aCount]:=Cleaned;
      inc(aCount);
      end;
    Cleaned:='';
  end;

begin
  aCount:=0;
  Cleaned:='';
  Result:=[];
  SetLength(Result,Length(aText) div MinWordLength);
  for C in aText do
    if Upcase(C) in ['A'..'Z'] then
      Cleaned:=Cleaned+C
    else
      MaybeAdd;
  MaybeAdd;
  SetLength(Result,aCount);
end;

procedure TPDFIndexer.CreateTables;
begin
  Connect;
  Indexer.CreateIndexerTables;
  DisConnect;
end;

procedure TPDFIndexer.IndexPDF(const aFileName: string);

Var
  F : TFileStream;


begin
  F:=TFileStream.Create(aFileName,fmOpenRead or fmShareDenyWrite);
  try
    if PDFURL='' then
      PDFURL:=ExtractFileName(aFileName);
    IndexPDF(F);
  finally
    F.Free;
  end;
end;

procedure TPDFIndexer.IndexPDF(const aStream: TStream);

begin
  DoLog('Start indexing PDF %s',[PDFURL]);
  Connect;
  try
    DoIndexPDF(aStream);
    DoLog('Done indexing PDF %s',[PDFURL]);
  finally
    Disconnect;
  end;
end;

end.


